@@ -64,6 +64,29 @@ VtkStructuredGrid = VtkFileType("StructuredGrid", ".vts")
 VtkUnstructuredGrid = VtkFileType("UnstructuredGrid", ".vtu")
 
 
+class VtkParallelFileType:
+    """
+    A wrapper class for parallel vtk file types.
+
+    Parameters
+    ----------
+    vtkftype : VtkFileType
+        Vtk file type
+    """
+
+    def __init__(self, vtkftype):
+        self.name = "P" + vtkftype.name
+        ext = vtkftype.ext
+        self.ext = ext[0] + "p" + ext[1:]
+
+
+VtkPImageData = VtkParallelFileType(VtkImageData)
+VtkPPolyData = VtkParallelFileType(VtkPolyData)
+VtkPRectilinearGrid = VtkParallelFileType(VtkRectilinearGrid)
+VtkPStructuredGrid = VtkParallelFileType(VtkStructuredGrid)
+VtkPUnstructuredGrid = VtkParallelFileType(VtkUnstructuredGrid)
+
+
 #    DATA TYPES
 class VtkDataType:
     """
@@ -323,9 +346,7 @@ class VtkFile:
             self.xml.addAttributes(Extent=ext)
 
         elif ncells and npoints:
-            self.xml.addAttributes(
-                NumberOfPoints=npoints, NumberOfCells=ncells
-            )
+            self.xml.addAttributes(NumberOfPoints=npoints, NumberOfCells=ncells)
 
         elif npoints or nverts or nlines or nstrips or npolys:
             if npoints is None:
@@ -512,9 +533,7 @@ class VtkFile:
         )
         self.xml.closeElement()
 
-        self.offset += (
-            nelem * ncomp * dtype.size + 8
-        )  # add 8 to indicate array size
+        self.offset += nelem * ncomp * dtype.size + 8  # add 8 to indicate array size
         return self
 
     def addData(self, name, data):
@@ -633,9 +652,9 @@ class VtkFile:
         from an external library.
         """
         if not self.appendedDataIsOpen:
-            self.xml.openElement("AppendedData").addAttributes(
-                encoding="raw"
-            ).addText("_")
+            self.xml.openElement("AppendedData").addAttributes(encoding="raw").addText(
+                "_"
+            )
             self.appendedDataIsOpen = True
 
     def closeAppendedData(self):
@@ -665,3 +684,237 @@ class VtkFile:
             self.xml.closeElement("AppendedData")
         self.xml.closeElement("VTKFile")
         self.xml.close()
+
+
+# ================================
+#        VtkParallelFile class
+# ================================
+class VtkParallelFile:
+    """
+    Class for a VTK parallel file.
+
+    Parameters
+    ----------
+    filepath : str
+        filename without extension
+    ftype : VtkParallelFileType
+    """
+
+    def __init__(self, filepath, ftype):
+        assert isinstance(ftype, VtkParallelFileType)
+        self.ftype = ftype
+        self.filename = filepath + ftype.ext
+        self.xml = XmlWriter(self.filename)
+        self.xml.openElement("VTKFile").addAttributes(
+            type=ftype.name,
+            version="1.0",
+            byte_order=_get_byte_order(),
+            header_type="UInt64",
+        )
+
+    def getFileName(self):
+        """Return absolute path to this file."""
+        return os.path.abspath(self.filename)
+
+    def addPiece(
+        self,
+        start=None,
+        end=None,
+        source=None,
+    ):
+        """
+        Add piece section with extent and source.
+        Parameters
+        ----------
+        start : array-like, optional
+            array or list with start indexes in each direction.
+            Must be given with end.
+        end : array-like, optional
+            array or list with end indexes in each direction.
+            Must be given with start.
+        source : str
+            Source of this piece
+        Returns
+        -------
+        VtkParallelFile
+            This VtkFile to allow chained calls.
+        """
+        # Check Source
+        assert source is not None
+        assert source.split(".")[-1] == self.ftype.ext[2:]
+
+        self.xml.openElement("Piece")
+        if start and end:
+            ext = _mix_extents(start, end)
+            self.xml.addAttributes(Extent=ext)
+        self.xml.addAttributes(Source=source)
+        self.xml.closeElement()
+        return self
+
+    def openData(
+        self,
+        nodeType,
+        scalars=None,
+        vectors=None,
+        normals=None,
+        tensors=None,
+        tcoords=None,
+    ):
+        """
+        Open data section.
+        Parameters
+        ----------
+        nodeType : str
+            Either "Point", "Cell" or "Field".
+        scalars : str, optional
+            default data array name for scalar data.
+        vectors : str, optional
+            default data array name for vector data.
+        normals : str, optional
+            default data array name for normals data.
+        tensors : str, optional
+            default data array name for tensors data.
+        tcoords : str, optional
+            default data array name for tcoords data.
+        Returns
+        -------
+        VtkFile
+            This VtkFile to allow chained calls.
+        """
+        self.xml.openElement(nodeType + "Data")
+        if scalars:
+            self.xml.addAttributes(Scalars=scalars)
+        if vectors:
+            self.xml.addAttributes(Vectors=vectors)
+        if normals:
+            self.xml.addAttributes(Normals=normals)
+        if tensors:
+            self.xml.addAttributes(Tensors=tensors)
+        if tcoords:
+            self.xml.addAttributes(TCoords=tcoords)
+
+        return self
+
+    def closeData(self, nodeType):
+        """
+        Close data section.
+        Parameters
+        ----------
+        nodeType : str
+            "Point", "Cell" or "Field".
+        Returns
+        -------
+        VtkFile
+            This VtkFile to allow chained calls.
+        """
+        self.xml.closeElement(nodeType + "Data")
+
+    def openGrid(self, start=None, end=None, origin=None, spacing=None, ghostlevel=0):
+        """
+        Open grid section.
+
+        Parameters
+        ----------
+        start : array-like, optional
+            array or list of start indexes.
+            Required for Structured, Rectilinear and ImageData grids.
+            The default is None.
+        end : array-like, optional
+            array or list of end indexes.
+            Required for Structured, Rectilinear and ImageData grids.
+            The default is None.
+        origin : array-like, optional
+            3D array or list with grid origin.
+            Only required for ImageData grids.
+            The default is None.
+        spacing : array-like, optional
+            3D array or list with grid spacing.
+            Only required for ImageData grids.
+            The default is None.
+        ghostlevel : int
+            Number of ghost-levels by which
+            the extents in the individual pieces overlap.
+        Returns
+        -------
+        VtkFile
+            This VtkFile to allow chained calls.
+        """
+        gType = self.ftype.name
+        self.xml.openElement(gType)
+
+        if gType == VtkPImageData.name:
+            if not start or not end or not origin or not spacing:
+                raise ValueError(f"start, end, origin and spacing required for {gType}")
+            ext = _mix_extents(start, end)
+            self.xml.addAttributes(
+                WholeExtent=ext,
+                Origin=_array_to_string(origin),
+                Spacing=_array_to_string(spacing),
+            )
+
+        elif gType in [VtkPStructuredGrid.name, VtkPRectilinearGrid.name]:
+            if not start or not end:
+                raise ValueError(f"start and end required for {gType}.")
+            ext = _mix_extents(start, end)
+            self.xml.addAttributes(WholeExtent=ext)
+
+        # Ghostlevel
+        self.xml.addAttributes(Ghostlevel=ghostlevel)
+        return self
+
+    def closeGrid(self):
+        """
+        Close grid element.
+        Returns
+        -------
+        VtkFile
+            This VtkFile to allow chained calls.
+        """
+        self.xml.closeElement(self.ftype.name)
+
+    def addHeader(self, name, dtype, ncomp):
+        """
+        Add data array description to xml header section.
+        Parameters
+        ----------
+        name : str
+            data array name.
+        dtype : str
+            data type.
+        ncomp : int
+            number of components, 1 (=scalar) and 3 (=vector).
+        Returns
+        -------
+
+        VtkFile
+            This VtkFile to allow chained calls.
+        Notes
+        -----
+
+        This is a low level function.
+        Use addData if you want to add a numpy array.
+        """
+        dtype = np_to_vtk[dtype.name]
+
+        self.xml.openElement("DataArray")
+        self.xml.addAttributes(
+            Name=name,
+            NumberOfComponents=ncomp,
+            type=dtype.name,
+        )
+        self.xml.closeElement()
+
+    def openElement(self, tagName):
+        """
+        Open an element.
+        Useful to add elements such as: Coordinates, Points, Verts, etc.
+        """
+        self.xml.openElement(tagName)
+
+    def closeElement(self, tagName):
+        self.xml.closeElement(tagName)
+
+    def save(self):
+        """Close file."""
+        self.xml.closeElement("VTKFile")
+        self.xml.close()
